import ncp from 'node-clipboardy';

/**
 * 数组转链表
 * @param initArr 原始数组
 * @returns
 */
function arrToLinkedList(initArr) {
    if (!Array.isArray(initArr) || initArr.length === 0) {
        return null;
    }
    let head = null;
    let tail = null;
    initArr.forEach((item) => {
        let node = {
            val: item,
            next: null,
        };
        if (head === null) {
            head = node;
            tail = node;
        }
        else {
            tail.next = node;
            tail = node;
        }
    });
    return head;
}

/**
 * 数组转双向链表
 * @param arr 数据源
 * @returns
 */
function arrToDoubleLinkedList(arr) {
    if (!Array.isArray(arr) || arr.length === 0) {
        return { head: null, tail: null };
    }
    let head = null;
    let tail = null;
    arr.forEach((item) => {
        let node = {
            val: item,
            next: null,
            prev: null,
        };
        if (head === null) {
            head = node;
            tail = node;
        }
        else {
            tail.next = node;
            node.prev = tail;
            tail = node;
        }
    });
    head.prev = tail;
    tail.next = head;
    return { head, tail };
}

/**
 * 数组转二叉树
 * @param  arr
 */
function arrToBinaryTree(arr) {
    if (!Array.isArray(arr) || arr.length === 0) {
        return null;
    }
    let root = null;
    let offset = 0;
    // 根节点从1开始
    let size = 1;
    let parentNodes = [];
    // 每次提取一层数据进行构建
    let levelChunk = arr.slice(offset, offset + size);
    // 用来存储每层构建的结果
    let buildChunk = [];
    offset += size;
    while (levelChunk.length) {
        let counter = 0;
        for (let i = 0; i < levelChunk.length; i++) {
            let val = levelChunk[i];
            let treeNode = null;
            if (val !== null) {
                // 计算出每层节点数
                counter++;
                treeNode = {
                    val,
                    left: null,
                    right: null,
                };
            }
            // 根据当前层节点的数据计算出父节点所在的位置
            let parentIdx = Math.floor(i / 2);
            let isLeftChild = i % 2 === 0;
            if (!parentNodes[parentIdx] && root === null) {
                root = treeNode;
            }
            else if (parentNodes[parentIdx]) {
                // 挂载左右儿子节点
                if (isLeftChild) {
                    parentNodes[parentIdx].left = treeNode;
                }
                else {
                    parentNodes[parentIdx].right = treeNode;
                }
            }
            buildChunk[i] = treeNode;
        }
        // 把当前层的空节点过滤掉，重新构建关系
        parentNodes = buildChunk.filter((x) => x !== null);
        // 计算出下一层的节点数
        size = 2 * counter;
        levelChunk = arr.slice(offset, offset + size);
        offset += size;
    }
    return root;
}

const template = `
class SimpleDsu {
  protected _set!: number[];

  public get dataSet() {
    return this._set;
  }

  constructor(n: number) {
    this.init(n);
  }

  /**
   * 初始化并查集
   * @param n
   */
  public init(n: number) {
    this._set = Array.from({
      length: n,
    }).fill(-1) as number[];
  }

  /**
   * 查找num在集合中是否存在，若存在返回其所在集合的跟节点的索引（即编号），若不存在，则返回-1
   * @param num
   * @returns
   */
  public find(num: number) {
    let idx = num - 1;
    // 查找并查集中不存在的元素
    if (idx < 0 || idx >= this._set.length) {
      return -1;
    }
    // 向上迭代寻找父元素
    while (this._set[idx] >= 0) {
      idx = this._set[idx];
    }
    return idx;
  }

  /**
   * 合并两个集合，按秩归并
   * @param num1
   * @param num2
   */
  public union(num1: number, num2: number) {
    let idx1 = this.find(num1);
    let idx2 = this.find(num2);
    // 说明大家不在同一个集合内
    if (idx1 != idx2) {
      // idx1的秩小于idx2的秩
      if (this._set[idx1] > this._set[idx2]) {
        // 将idx1的跟节点指向idx2的根节点
        this._set[idx1] = idx2;
      } else {
        if (this._set[idx1] === this._set[idx2]) {
          this._set[idx1]--;
        }
        this._set[idx2] = idx1;
      }
    }
  }

  /**
   * 统计集合的个数
   * @returns
   */
  public count() {
    return this._set.reduce((total, item) => {
      return total + (item < 0 ? 1 : 0);
    }, 0);
  }
}
`;
/**
 * 简单版本的并查集实现，即数组的索引即代表的是它所存储的数据
 */
class SimpleDsu {
    get dataSet() {
        return this._set;
    }
    constructor(n) {
        this.init(n);
        ncp.writeSync(template);
    }
    /**
     * 初始化并查集
     * @param n
     */
    init(n) {
        this._set = Array.from({
            length: n,
        }).fill(-1);
    }
    /**
     * 查找num在集合中是否存在，若存在返回其所在集合的跟节点的索引（即编号），若不存在，则返回-1
     * @param num
     * @returns
     */
    find(num) {
        let idx = num - 1;
        // 查找并查集中不存在的元素
        if (idx < 0 || idx >= this._set.length) {
            return -1;
        }
        // 向上迭代寻找父元素
        while (this._set[idx] >= 0) {
            idx = this._set[idx];
        }
        return idx;
    }
    /**
     * 合并两个集合，按秩归并
     * @param num1
     * @param num2
     */
    union(num1, num2) {
        let idx1 = this.find(num1);
        let idx2 = this.find(num2);
        // 说明大家不在同一个集合内
        if (idx1 != idx2) {
            // idx1的秩小于idx2的秩
            if (this._set[idx1] > this._set[idx2]) {
                // 将idx1的跟节点指向idx2的根节点
                this._set[idx1] = idx2;
            }
            else {
                if (this._set[idx1] === this._set[idx2]) {
                    this._set[idx1]--;
                }
                this._set[idx2] = idx1;
            }
        }
    }
    /**
     * 统计集合的个数
     * @returns
     */
    count() {
        return this._set.reduce((total, item) => {
            return total + (item < 0 ? 1 : 0);
        }, 0);
    }
}

class GeneralDsu {
    constructor() {
        /**
         * 集合存储域
         */
        this._set = [];
        /**
         * 判等依据
         * @param data
         * @param target
         * @returns
         */
        this._equalCondition = (data, target) => {
            return data === target;
        };
        const template = `
interface GeneralDsuElement<T> {
  /**
   * 数据域
   */
  dat: T;
  /**
   * 父节点的索引，如果为负数，则没有父节点
   */
  parent: number;
}

class GeneralDsu<T> {
  /**
   * 集合存储域
   */
  private _set: GeneralDsuElement<T>[] = [];

  /**
   * 外部访问的集合
   */
  public get dataSet() {
    return this._set;
  }

  /**
   * 判等依据
   * @param data
   * @param target
   * @returns
   */
  private _equalCondition = (data: T, target: unknown) => {
    return data === target;
  };

  /**
   * 设置判等条件
   * @param outerEqualCondition
   */
  public setEqual(outerEqualCondition: (data: T, target: unknown) => boolean) {
    this._equalCondition = outerEqualCondition;
  }

  /**
   * 初始化并查集
   * @param values
   */
  init(values: T[]) {
    values.forEach((v) => {
      this._set.push({
        // 初始化的时候，每个子树只有一个元素
        parent: -1,
        dat: v,
      });
    });
  }

  /**
   * 查找元素是否在集合中，若存在，则返回根节点所在的索引，若不在，则返回-1
   * @param target 目标元素
   */
  find(target: unknown): number {
    for (let i = 0; i < this._set.length; i++) {
      // 元素在数组中能够被找到
      if (this._equalCondition(this._set[i].dat, target)) {
        let pos = i;
        // 尝试找这个元素的祖先节点
        while (this._set[pos].parent >= 0) {
          // 继续向上查找根节点
          pos = this._set[pos].parent;
        }
        // 路径压缩：将路径上的节点都直接连接到根节点
        let current = i;
        while (current !== pos) {
          const next = this._set[current].parent;
          this._set[current].parent = pos;
          current = next;
        }
        return pos;
      }
    }
    // 未找到
    return -1;
  }

  /**
   * 合并集合
   * @param set1
   * @param set2
   */
  union(set1: T, set2: T) {
    const r1 = this.find(set1);
    const r2 = this.find(set2);
    if (r1 != r2) {
      const val1 = Math.abs(this._set[r1].parent);
      const val2 = Math.abs(this._set[r2].parent);
      // 将小树贴到大树上
      if (val1 < val2) {
        this._set[r1].parent = r2;
      } else {
        // 如果两棵树相同，则需要将树的规模增加
        if (val1 === val2) {
          this._set[r1].parent--;
        }
        this._set[r2].parent = r1;
      }
    }
  }

  /**
   * 统计并查集中子树的个数
   * @returns
   */
  count() {
    let size = 0;
    this._set.forEach((v) => {
      if (v.parent < 0) {
        size++;
      }
    });
    return size;
  }
}
`;
        ncp.writeSync(template);
    }
    /**
     * 外部访问的集合
     */
    get dataSet() {
        return this._set;
    }
    /**
     * 设置判等条件
     * @param outerEqualCondition
     */
    setEqual(outerEqualCondition) {
        this._equalCondition = outerEqualCondition;
    }
    /**
     * 初始化并查集
     * @param values
     */
    init(values) {
        values.forEach((v) => {
            this._set.push({
                // 初始化的时候，每个子树只有一个元素
                parent: -1,
                dat: v,
            });
        });
    }
    /**
     * 查找元素是否在集合中，若存在，则返回根节点所在的索引，若不在，则返回-1
     * @param target 目标元素
     */
    find(target) {
        for (let i = 0; i < this._set.length; i++) {
            // 元素在数组中能够被找到
            if (this._equalCondition(this._set[i].dat, target)) {
                let pos = i;
                // 尝试找这个元素的祖先节点
                while (this._set[pos].parent >= 0) {
                    // 继续向上查找根节点
                    pos = this._set[pos].parent;
                }
                // 路径压缩：将路径上的节点都直接连接到根节点
                let current = i;
                while (current !== pos) {
                    const next = this._set[current].parent;
                    this._set[current].parent = pos;
                    current = next;
                }
                return pos;
            }
        }
        // 未找到
        return -1;
    }
    /**
     * 合并集合
     * @param set1
     * @param set2
     */
    union(set1, set2) {
        const r1 = this.find(set1);
        const r2 = this.find(set2);
        if (r1 != r2) {
            const val1 = Math.abs(this._set[r1].parent);
            const val2 = Math.abs(this._set[r2].parent);
            // 将小树贴到大树上
            if (val1 < val2) {
                this._set[r1].parent = r2;
            }
            else {
                // 如果两棵树相同，则需要将树的规模增加
                if (val1 === val2) {
                    this._set[r1].parent--;
                }
                this._set[r2].parent = r1;
            }
        }
    }
    /**
     * 统计并查集中子树的个数
     * @returns
     */
    count() {
        let size = 0;
        this._set.forEach((v) => {
            if (v.parent < 0) {
                size++;
            }
        });
        return size;
    }
}

/*
 * 抽象堆
 */
class Heap {
    get count() {
        return this._data.length;
    }
    constructor(...initElements) {
        /**
         * 定义一个存储数据的内存空间
         */
        this._data = [];
        this._selfCompare = (compareVal, currentVal) => {
            if (typeof this._compare !== "function") {
                throw "cannot compare without a compare function";
            }
            return this._compare(compareVal, currentVal);
        };
        this._data.push(...initElements);
        this.writeIntoClipboard();
    }
    /**
     * 设置比较函数
     * @param {(compareVal, currentVal) => boolean} compareFunc
     */
    setCompare(compareFunc) {
        this._compare = compareFunc;
    }
    /**
     * 获取堆顶的元素
     */
    getTop() {
        return this._data[0];
    }
    /**
     * 按断堆是否是空
     */
    isEmpty() {
        return this._data.length === 0;
    }
    /**
     * 向堆中插入一个合法值
     * @param  val
     */
    insertQueue(val) {
        // 如果当前数组没有元素，直接插入即可
        if (this._data.length === 0) {
            this._data[0] = val;
        }
        else {
            // 让i指向当前新位置，因为没有哨兵的关系，最后一个元素是length - 1，所以新位置就是length
            let i = this.count;
            while (this._selfCompare(val, this._data[Math.floor(i / 2)]) && i > 0) {
                this._data[i] = this._data[Math.floor(i / 2)];
                i = Math.floor(i / 2);
            }
            // 在合适的位置插入新的值
            this._data[i] = val;
        }
    }
    /**
     * 获取堆中的最值元素
     * @returns
     */
    deleteTop() {
        if (this.isEmpty()) {
            throw "can not delete element from empty heap";
        }
        // 取出堆顶的元素
        let minVal = this._data[0];
        let temp = this._data[this.count - 1];
        this._data[0] = temp;
        // JavaScript语言需要进行这一步，让数组的规模缩小，释放空间
        this._data.length--;
        // 如果当前堆里面还存在元素的话，将
        this.percDown(0);
        return minVal;
    }
    /**
     * 下滤：将堆中以堆data[p]为根的子堆调整为最小堆
     * @param  p 根节点索引
     */
    percDown(p) {
        // 如果当前堆元素小于1个，就不执行调整操作
        if (this.count <= 1) {
            return;
        }
        let parent, child;
        /* 取出根结点存放的值 */
        let temp = this._data[p];
        for (parent = p; parent * 2 < this.count; parent = child) {
            child = parent * 2;
            if (child + 1 < this.count && this._selfCompare(this._data[child + 1], this._data[child])) {
                child++; /* child指向左右子结点的较大者 最大堆 较小者 最小堆 */
            }
            /* 找到了合适位置 */
            // 注意这儿一定要取得等号 temp <= this.#data[child] 最小堆 temp >= this.#data[child] 最大堆
            if (this._selfCompare(temp, this._data[child])) {
                break;
            }
            else {
                /* 下滤X */
                this._data[parent] = this._data[child];
            }
        }
        this._data[parent] = temp;
    }
    /**
     * 构建最小堆
     */
    buildHeap() {
        /* 调整data中的元素，使满足最小堆的有序性  */
        /* 这里所有size个元素已经存在data[]中 */
        /* 从最后一个结点的父节点开始，到根结点1 */
        for (let i = Math.floor((this.count - 1) / 2); i >= 0; i--) {
            this.percDown(i);
        }
    }
    writeIntoClipboard() {
        const template = `
    class Heap<T> {
      /**
       * 定义一个存储数据的内存空间
       */
      private _data: T[] = [];
    
      /**
       * 比较函数
       */
      private _compare!: (compareVal: T, currentVal: T) => boolean;
    
      public get count() {
        return this._data.length;
      }
    
      constructor(...initElements: T[]) {
        this._data.push(...initElements);
      }
    
      private _selfCompare = (compareVal: T, currentVal: T) => {
        if (typeof this._compare !== "function") {
          throw "cannot compare without a compare function";
        }
        return this._compare(compareVal, currentVal);
      };
    
      /**
       * 设置比较函数
       * @param {(compareVal, currentVal) => boolean} compareFunc
       */
      public setCompare(compareFunc: (compareVal: T, currentVal: T) => boolean) {
        this._compare = compareFunc;
      }
    
      /**
       * 获取堆顶的元素
       */
      protected getTop() {
        return this._data[0];
      }
    
      /**
       * 按断堆是否是空
       */
      public isEmpty() {
        return this._data.length === 0;
      }
    
      /**
       * 向堆中插入一个合法值
       * @param  val
       */
      public insertQueue(val: T) {
        // 如果当前数组没有元素，直接插入即可
        if (this._data.length === 0) {
          this._data[0] = val;
        } else {
          // 让i指向当前新位置，因为没有哨兵的关系，最后一个元素是length - 1，所以新位置就是length
          let i = this.count;
          while (this._selfCompare(val, this._data[Math.floor(i / 2)]) && i > 0) {
            this._data[i] = this._data[Math.floor(i / 2)];
            i = Math.floor(i / 2);
          }
          // 在合适的位置插入新的值
          this._data[i] = val;
        }
      }
    
      /**
       * 获取堆中的最值元素
       * @returns
       */
      public deleteTop() {
        if (this.isEmpty()) {
          throw "can not delete element from empty heap";
        }
        // 取出堆顶的元素
        let minVal = this._data[0];
        let temp = this._data[this.count - 1];
        this._data[0] = temp;
        // JavaScript语言需要进行这一步，让数组的规模缩小，释放空间
        this._data.length--;
        // 如果当前堆里面还存在元素的话，将
        this.percDown(0);
        return minVal;
      }
    
      /**
       * 下滤：将堆中以堆data[p]为根的子堆调整为最小堆
       * @param  p 根节点索引
       */
      private percDown(p: number) {
        // 如果当前堆元素小于1个，就不执行调整操作
        if (this.count <= 1) {
          return;
        }
        let parent, child;
        /* 取出根结点存放的值 */
        let temp = this._data[p];
        for (parent = p; parent * 2 < this.count; parent = child) {
          child = parent * 2;
          if (child + 1 < this.count && this._selfCompare(this._data[child + 1], this._data[child])) {
            child++; /* child指向左右子结点的较大者 最大堆 较小者 最小堆 */
          }
          /* 找到了合适位置 */
          // 注意这儿一定要取得等号 temp <= this.#data[child] 最小堆 temp >= this.#data[child] 最大堆
          if (this._selfCompare(temp, this._data[child])) {
            break;
          } else {
            /* 下滤X */
            this._data[parent] = this._data[child];
          }
        }
        this._data[parent] = temp;
      }
    
      /**
       * 构建最小堆
       */
      public buildHeap() {
        /* 调整data中的元素，使满足最小堆的有序性  */
        /* 这里所有size个元素已经存在data[]中 */
        /* 从最后一个结点的父节点开始，到根结点1 */
        for (let i = Math.floor((this.count - 1) / 2); i >= 0; i--) {
          this.percDown(i);
        }
      }
    }
    `;
        ncp.writeSync(template);
    }
}

class SimpleMaxHeap extends Heap {
    constructor(...initElements) {
        super(...initElements);
        // 比较依据，若targetVal比目标节点大，则需要提前
        this.setCompare((targetVal, compareVal) => {
            return targetVal >= compareVal;
        });
        this.buildHeap();
    }
    getMax() {
        return this.getTop();
    }
    deleteMax() {
        return this.deleteTop();
    }
    writeIntoClipboard() {
        const template = `
    class Heap<T> {
      /**
       * 定义一个存储数据的内存空间
       */
      private _data: T[] = [];
    
      /**
       * 比较函数
       */
      private _compare!: (compareVal: T, currentVal: T) => boolean;
    
      public get count() {
        return this._data.length;
      }
    
      constructor(...initElements: T[]) {
        this._data.push(...initElements);
      }
    
      private _selfCompare = (compareVal: T, currentVal: T) => {
        if (typeof this._compare !== "function") {
          throw "cannot compare without a compare function";
        }
        return this._compare(compareVal, currentVal);
      };
    
      /**
       * 设置比较函数
       * @param {(compareVal, currentVal) => boolean} compareFunc
       */
      public setCompare(compareFunc: (compareVal: T, currentVal: T) => boolean) {
        this._compare = compareFunc;
      }
    
      /**
       * 获取堆顶的元素
       */
      protected getTop() {
        return this._data[0];
      }
    
      /**
       * 按断堆是否是空
       */
      public isEmpty() {
        return this._data.length === 0;
      }
    
      /**
       * 向堆中插入一个合法值
       * @param  val
       */
      public insertQueue(val: T) {
        // 如果当前数组没有元素，直接插入即可
        if (this._data.length === 0) {
          this._data[0] = val;
        } else {
          // 让i指向当前新位置，因为没有哨兵的关系，最后一个元素是length - 1，所以新位置就是length
          let i = this.count;
          while (this._selfCompare(val, this._data[Math.floor(i / 2)]) && i > 0) {
            this._data[i] = this._data[Math.floor(i / 2)];
            i = Math.floor(i / 2);
          }
          // 在合适的位置插入新的值
          this._data[i] = val;
        }
      }
    
      /**
       * 获取堆中的最值元素
       * @returns
       */
      public deleteTop() {
        if (this.isEmpty()) {
          throw "can not delete element from empty heap";
        }
        // 取出堆顶的元素
        let minVal = this._data[0];
        let temp = this._data[this.count - 1];
        this._data[0] = temp;
        // JavaScript语言需要进行这一步，让数组的规模缩小，释放空间
        this._data.length--;
        // 如果当前堆里面还存在元素的话，将
        this.percDown(0);
        return minVal;
      }
    
      /**
       * 下滤：将堆中以堆data[p]为根的子堆调整为最小堆
       * @param  p 根节点索引
       */
      private percDown(p: number) {
        // 如果当前堆元素小于1个，就不执行调整操作
        if (this.count <= 1) {
          return;
        }
        let parent, child;
        /* 取出根结点存放的值 */
        let temp = this._data[p];
        for (parent = p; parent * 2 < this.count; parent = child) {
          child = parent * 2;
          if (child + 1 < this.count && this._selfCompare(this._data[child + 1], this._data[child])) {
            child++; /* child指向左右子结点的较大者 最大堆 较小者 最小堆 */
          }
          /* 找到了合适位置 */
          // 注意这儿一定要取得等号 temp <= this.#data[child] 最小堆 temp >= this.#data[child] 最大堆
          if (this._selfCompare(temp, this._data[child])) {
            break;
          } else {
            /* 下滤X */
            this._data[parent] = this._data[child];
          }
        }
        this._data[parent] = temp;
      }
    
      /**
       * 构建最小堆
       */
      public buildHeap() {
        /* 调整data中的元素，使满足最小堆的有序性  */
        /* 这里所有size个元素已经存在data[]中 */
        /* 从最后一个结点的父节点开始，到根结点1 */
        for (let i = Math.floor((this.count - 1) / 2); i >= 0; i--) {
          this.percDown(i);
        }
      }
    }

    class SimpleMaxHeap extends Heap<number> {
      constructor(...initElements: number[]) {
        super(...initElements);
        // 比较依据，若targetVal比目标节点大，则需要提前
        this.setCompare((targetVal, compareVal) => {
          return targetVal >= compareVal;
        });
        this.buildHeap();
      }

      getMax() {
        return this.getTop();
      }

      deleteMax() {
        return this.deleteTop();
      }
    }
    `;
        ncp.writeSync(template);
    }
}

class SimpleMinHeap extends Heap {
    constructor(...initElements) {
        super(...initElements);
        // 比较依据，若targetVal比目标节点大，则需要下滤
        this.setCompare((targetVal, compareVal) => {
            return targetVal <= compareVal;
        });
        this.buildHeap();
    }
    getMin() {
        return this.getTop();
    }
    deleteMin() {
        return this.deleteTop();
    }
    writeIntoClipboard() {
        const template = `
  class Heap<T> {
    /**
     * 定义一个存储数据的内存空间
     */
    private _data: T[] = [];
  
    /**
     * 比较函数
     */
    private _compare!: (compareVal: T, currentVal: T) => boolean;
  
    public get count() {
      return this._data.length;
    }
  
    constructor(...initElements: T[]) {
      this._data.push(...initElements);
    }
  
    private _selfCompare = (compareVal: T, currentVal: T) => {
      if (typeof this._compare !== "function") {
        throw "cannot compare without a compare function";
      }
      return this._compare(compareVal, currentVal);
    };
  
    /**
     * 设置比较函数
     * @param {(compareVal, currentVal) => boolean} compareFunc
     */
    public setCompare(compareFunc: (compareVal: T, currentVal: T) => boolean) {
      this._compare = compareFunc;
    }
  
    /**
     * 获取堆顶的元素
     */
    protected getTop() {
      return this._data[0];
    }
  
    /**
     * 按断堆是否是空
     */
    public isEmpty() {
      return this._data.length === 0;
    }
  
    /**
     * 向堆中插入一个合法值
     * @param  val
     */
    public insertQueue(val: T) {
      // 如果当前数组没有元素，直接插入即可
      if (this._data.length === 0) {
        this._data[0] = val;
      } else {
        // 让i指向当前新位置，因为没有哨兵的关系，最后一个元素是length - 1，所以新位置就是length
        let i = this.count;
        while (this._selfCompare(val, this._data[Math.floor(i / 2)]) && i > 0) {
          this._data[i] = this._data[Math.floor(i / 2)];
          i = Math.floor(i / 2);
        }
        // 在合适的位置插入新的值
        this._data[i] = val;
      }
    }
  
    /**
     * 获取堆中的最值元素
     * @returns
     */
    public deleteTop() {
      if (this.isEmpty()) {
        throw "can not delete element from empty heap";
      }
      // 取出堆顶的元素
      let minVal = this._data[0];
      let temp = this._data[this.count - 1];
      this._data[0] = temp;
      // JavaScript语言需要进行这一步，让数组的规模缩小，释放空间
      this._data.length--;
      // 如果当前堆里面还存在元素的话，将
      this.percDown(0);
      return minVal;
    }
  
    /**
     * 下滤：将堆中以堆data[p]为根的子堆调整为最小堆
     * @param  p 根节点索引
     */
    private percDown(p: number) {
      // 如果当前堆元素小于1个，就不执行调整操作
      if (this.count <= 1) {
        return;
      }
      let parent, child;
      /* 取出根结点存放的值 */
      let temp = this._data[p];
      for (parent = p; parent * 2 < this.count; parent = child) {
        child = parent * 2;
        if (child + 1 < this.count && this._selfCompare(this._data[child + 1], this._data[child])) {
          child++; /* child指向左右子结点的较大者 最大堆 较小者 最小堆 */
        }
        /* 找到了合适位置 */
        // 注意这儿一定要取得等号 temp <= this.#data[child] 最小堆 temp >= this.#data[child] 最大堆
        if (this._selfCompare(temp, this._data[child])) {
          break;
        } else {
          /* 下滤X */
          this._data[parent] = this._data[child];
        }
      }
      this._data[parent] = temp;
    }
  
    /**
     * 构建最小堆
     */
    public buildHeap() {
      /* 调整data中的元素，使满足最小堆的有序性  */
      /* 这里所有size个元素已经存在data[]中 */
      /* 从最后一个结点的父节点开始，到根结点1 */
      for (let i = Math.floor((this.count - 1) / 2); i >= 0; i--) {
        this.percDown(i);
      }
    }
  }

  class SimpleMinHeap extends Heap<number> {
    constructor(...initElements: number[]) {
      super(...initElements);
      // 比较依据，若targetVal比目标节点大，则需要下滤
      this.setCompare((targetVal, compareVal) => {
        return targetVal <= compareVal;
      });
      this.buildHeap();
    }
  
    getMin() {
      return this.getTop();
    }
  
    deleteMin() {
      return this.deleteTop();
    }
  }
  `;
        ncp.writeSync(template);
    }
}

const generalTpl = `
/**
 * 合并两个有序数组
 * @param arr1 有序数组1
 * @param arr2 有序数组2
 * @param mergePredicate 合并依据
 * @returns
 */
function merge<T>(arr1: T[], arr2: T[], mergePredicate: (val1: T, val2: T) => boolean): T[] {
  const mergedResult: T[] = [];
  let offset1 = 0;
  let offset2 = 0;
  while (offset1 < arr1.length && offset2 < arr2.length) {
    const val1 = arr1[offset1];
    const val2 = arr2[offset2];
    if (mergePredicate(val1, val2)) {
      mergedResult.push(val1);
      offset1++;
    } else {
      mergedResult.push(val2);
      offset2++;
    }
  }

  while (offset1 < arr1.length) {
    const val1 = arr1[offset1++];
    mergedResult.push(val1);
  }

  while (offset2 < arr2.length) {
    const val2 = arr2[offset2++];
    mergedResult.push(val2);
  }

  return mergedResult;
}
`;
const ascTpl = `
/**
 * 合并两个有序数组
 * @param arr1 有序数组1
 * @param arr2 有序数组2
 * @param mergePredicate 合并依据
 * @returns
 */
function merge<T>(arr1: T[], arr2: T[], mergePredicate: (val1: T, val2: T) => boolean): T[] {
  const mergedResult: T[] = [];
  let offset1 = 0;
  let offset2 = 0;
  while (offset1 < arr1.length && offset2 < arr2.length) {
    const val1 = arr1[offset1];
    const val2 = arr2[offset2];
    if (mergePredicate(val1, val2)) {
      mergedResult.push(val1);
      offset1++;
    } else {
      mergedResult.push(val2);
      offset2++;
    }
  }

  while (offset1 < arr1.length) {
    const val1 = arr1[offset1++];
    mergedResult.push(val1);
  }

  while (offset2 < arr2.length) {
    const val2 = arr2[offset2++];
    mergedResult.push(val2);
  }

  return mergedResult;
}

/**
 * 升序合并两个有序数组
 * @param arr1
 * @param arr2
 * @returns
 */
function mergeByAsc(arr1: number[], arr2: number[]) {
  const mergePredicate = (a: number, b: number) => {
    return a < b;
  };
  return merge(arr1, arr2, mergePredicate);
}
`;
const descTpl = `
/**
 * 合并两个有序数组
 * @param arr1 有序数组1
 * @param arr2 有序数组2
 * @param mergePredicate 合并依据
 * @returns
 */
function merge<T>(arr1: T[], arr2: T[], mergePredicate: (val1: T, val2: T) => boolean): T[] {
  const mergedResult: T[] = [];
  let offset1 = 0;
  let offset2 = 0;
  while (offset1 < arr1.length && offset2 < arr2.length) {
    const val1 = arr1[offset1];
    const val2 = arr2[offset2];
    if (mergePredicate(val1, val2)) {
      mergedResult.push(val1);
      offset1++;
    } else {
      mergedResult.push(val2);
      offset2++;
    }
  }

  while (offset1 < arr1.length) {
    const val1 = arr1[offset1++];
    mergedResult.push(val1);
  }

  while (offset2 < arr2.length) {
    const val2 = arr2[offset2++];
    mergedResult.push(val2);
  }

  return mergedResult;
}

/**
 * 降序合并两个有序数组
 * @param arr1
 * @param arr2
 * @returns
 */
function mergeByDesc(arr1: number[], arr2: number[]) {
  const mergePredicate = (a: number, b: number) => {
    return b < a;
  };
  return merge(arr1, arr2, mergePredicate);
}
`;
/**
 * 合并两个有序数组
 * @param arr1 有序数组1
 * @param arr2 有序数组2
 * @param mergePredicate 合并依据
 * @returns
 */
function merge(arr1, arr2, mergePredicate, writeSourceCode = () => {
    ncp.writeSync(generalTpl);
}) {
    const mergedResult = [];
    let offset1 = 0;
    let offset2 = 0;
    while (offset1 < arr1.length && offset2 < arr2.length) {
        const val1 = arr1[offset1];
        const val2 = arr2[offset2];
        if (mergePredicate(val1, val2)) {
            mergedResult.push(val1);
            offset1++;
        }
        else {
            mergedResult.push(val2);
            offset2++;
        }
    }
    while (offset1 < arr1.length) {
        const val1 = arr1[offset1++];
        mergedResult.push(val1);
    }
    while (offset2 < arr2.length) {
        const val2 = arr2[offset2++];
        mergedResult.push(val2);
    }
    writeSourceCode();
    return mergedResult;
}
/**
 * 升序合并两个有序数组
 * @param arr1
 * @param arr2
 * @returns
 */
function mergeByAsc(arr1, arr2) {
    const mergePredicate = (a, b) => {
        return a < b;
    };
    return merge(arr1, arr2, mergePredicate, () => {
        ncp.writeSync(ascTpl);
    });
}
/**
 * 降序合并两个有序数组
 * @param arr1
 * @param arr2
 * @returns
 */
function mergeByDesc(arr1, arr2) {
    const mergePredicate = (a, b) => {
        return b < a;
    };
    return merge(arr1, arr2, mergePredicate, () => {
        ncp.writeSync(descTpl);
    });
}

export { GeneralDsu, Heap, SimpleDsu, SimpleMaxHeap, SimpleMinHeap, arrToBinaryTree, arrToDoubleLinkedList, arrToLinkedList, merge, mergeByAsc, mergeByDesc };
